import dataclasses
from typing import Dict, List
from enova.common.config import CONFIG
from enova.api.base import ASyncAPI
from enova.common.constant import HttpMethod
from enova.common.logger import LOGGER
from enova.server.restful.serializer import EmergingAIBaseModel
from enova.common.error import PilotApiResponseError


PILOT_API_HOST = CONFIG.api["pilot_api_host"]


# ----------------


class GetPilotStatusRequest(EmergingAIBaseModel):
    task_name: str


class GetDeployRequest(EmergingAIBaseModel):
    task_name: str


class EnvKVPair(EmergingAIBaseModel):
    name: str
    value: str


class GPUSpec(EmergingAIBaseModel):
    name: str
    num: int
    spec: int


class LLMSpec(EmergingAIBaseModel):
    framework: str
    param: int


class ModelConfig(EmergingAIBaseModel):
    gpu: GPUSpec
    llm: LLMSpec
    version: str


class VolumePaths(EmergingAIBaseModel):
    hostPath: str
    mountPath: str


class CreatePilotDeployParameters(EmergingAIBaseModel):
    backend: str
    backendConfig: dict
    envs: List[EnvKVPair]
    exporter_endpoint: str
    exporter_service_name: str
    model: str
    modelConfig: ModelConfig
    name: str
    port: int
    replica: int
    volumes: List[VolumePaths]


class DeleteDeployRequest(EmergingAIBaseModel):
    task_name: str


class ListDetectHistoryRequest(EmergingAIBaseModel):
    task_name: str


# ----------------


class _PilotRequestFormer:
    def get_deploy(self, params) -> Dict:
        req_body = {"task_name": params["task_name"]}
        req_body = GetDeployRequest(**req_body)
        return req_body.dict()

    def create_deploy(self, params) -> Dict:
        req_body = params
        req_body = CreatePilotDeployParameters(**req_body)
        return req_body.dict()

    def delete_deploy(self, params) -> Dict:
        req_body = {"task_name": params["task_name"]}
        req_body = DeleteDeployRequest(**req_body)
        return req_body.dict()

    def list_detect_history(self, params) -> Dict:
        req_body = {"task_name": params["task_name"]}
        req_body = ListDetectHistoryRequest(**req_body)
        return req_body.dict()


class _PilotResponseReformer:
    def get_deploy(self, ret) -> Dict:
        if ret.get("code") and ret["code"] not in [0]:
            raise PilotApiResponseError()

        resp_dict = {"ret": ret}
        return resp_dict

    def create_deploy(self, ret) -> Dict:
        if ret.get("code") and ret["code"] not in [0]:
            raise PilotApiResponseError()

        resp_dict = {"ret": ret}
        return resp_dict

    def delete_deploy(self, ret) -> Dict:
        if ret.get("code") and ret["code"] not in [0]:
            raise PilotApiResponseError()

        resp_dict = {"ret": ret}
        return resp_dict

    def list_detect_history(self, ret) -> Dict:
        if ret.get("code") and ret["code"] not in [0]:
            raise PilotApiResponseError()

        resp_dict = {"ret": ret}
        return resp_dict


class _PilotApi:
    def __init__(self) -> None:
        # self.xxx = ASyncAPI()
        LOGGER.debug(f"pilot api init: PILOT_API_HOST={PILOT_API_HOST}")
        self.get_deploy = ASyncAPI(
            url=PILOT_API_HOST + "/api/pilot/v1/docker/deploy",
            method=HttpMethod.GET.value,
            header_builder=self.get_request_header,
        )
        self.create_deploy = ASyncAPI(
            url=PILOT_API_HOST + "/api/pilot/v1/docker/deploy",
            method=HttpMethod.POST.value,
            post_send=self.post_send,
            header_builder=self.get_request_header,
        )
        self.delete_deploy = ASyncAPI(
            url=PILOT_API_HOST + "/api/pilot/v1/docker/deploy",
            method=HttpMethod.DELETE.value,
            header_builder=self.get_request_header,
        )
        self.list_detect_history = ASyncAPI(
            url=PILOT_API_HOST + "/api/pilot/v1/task/detect/history",
            method=HttpMethod.GET.value,
            header_builder=self.get_request_header,
        )

    def post_send(self, ret: Dict):
        return ret

    def get_request_header(self):
        # TODO:
        return {}


@dataclasses.dataclass
class PilotApiWrapper:
    def __post_init__(self) -> None:
        self.async_api = _PilotApi()
        self.request_former = _PilotRequestFormer
        self.response_reformer = _PilotResponseReformer

    def __getattr__(self, item):
        if all(
            [
                hasattr(self.request_former, item),
                hasattr(self.async_api, item),
                hasattr(self.response_reformer, item),
            ]
        ):

            async def pipeline(params):
                req_body = getattr(self.request_former, item)(self, params)
                api_ret = await getattr(self.async_api, item)(req_body)
                res = getattr(self.response_reformer, item)(self, api_ret)

                return res

            return pipeline

        if hasattr(self.async_api, item):
            return getattr(self.async_api, item)
        return super().__getattr__(item)
